@@ -1,6 +1,5 @@
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TupleSections #-}
@@ -85,10 +84,14 @@ module Data.List.Infinite (
   stripPrefix,
 
   -- * Searching
+  filter,
   lookup,
   find,
-  filter,
+  mapMaybe,
+  catMaybes,
   partition,
+  mapEither,
+  partitionEithers,
 
   -- * Indexing
   (!!),
@@ -152,12 +155,14 @@ import Control.Monad (Monad (..))
 import Data.Bits ((.&.))
 import Data.Char (Char, isSpace)
 import Data.Coerce (coerce)
+import Data.Either (Either, either)
 import Data.Eq (Eq, (/=), (==))
 import qualified Data.Foldable as F
 import Data.Functor (Functor (..))
 import qualified Data.List as List
 import Data.List.NonEmpty (NonEmpty (..))
 import qualified Data.List.NonEmpty as NE
+import Data.Maybe (maybe)
 import Data.Ord (Ord, Ordering (..), compare, (<), (<=), (>), (>=))
 import qualified GHC.Exts
 import Numeric.Natural (Natural)
@@ -173,7 +178,8 @@ import Data.List.Infinite.Internal
 import Data.List.Infinite.Zip
 
 -- | Right-associative fold of an infinite list, necessarily lazy in the accumulator.
--- Any unconditional attempt to force the accumulator even to WHNF
+-- Any unconditional attempt to force the accumulator even
+-- to the weak head normal form (WHNF)
 -- will hang the computation. E. g., the following definition isn't productive:
 --
 -- > import Data.List.NonEmpty (NonEmpty(..))
@@ -182,6 +188,8 @@ import Data.List.Infinite.Zip
 -- One should use lazy patterns, e. g.,
 --
 -- > toNonEmpty = foldr (\a ~(x :| xs) -> a :| x : xs)
+--
+-- This is a catamorphism on infinite lists.
 foldr :: (a -> b -> b) -> Infinite a -> b
 foldr f = go
   where
@@ -197,7 +205,14 @@ foldr f = go
     cons x (g cons)
   #-}
 
--- | Convert to a list. Use 'cycle' to go in another direction.
+-- | Paramorphism on infinite lists.
+para :: forall a b. (a -> Infinite a -> b -> b) -> Infinite a -> b
+para f = go
+  where
+    go :: Infinite a -> b
+    go (x :< xs) = f x xs (go xs)
+
+-- | Convert to a list. Use 'cycle' to go in the opposite direction.
 toList :: Infinite a -> [a]
 toList = foldr (:)
 {-# NOINLINE [0] toList #-}
@@ -208,7 +223,7 @@ toList = foldr (:)
     GHC.Exts.build (\cons -> const (foldr cons xs))
   #-}
 
--- | Generate infinite sequences, starting from a given element,
+-- | Generate an infinite progression, starting from a given element,
 -- similar to @[x..]@.
 -- For better user experience consider enabling @{\-# LANGUAGE PostfixOperators #-\}@:
 --
@@ -221,10 +236,16 @@ toList = foldr (:)
 -- >>> :set -XPostfixOperators
 -- >>> Data.List.Infinite.take 10 (EQ...)
 -- [EQ,GT,EQ,GT,EQ,GT,EQ,GT,EQ,GT]
+--
+-- Remember that 'Int' is a finite type as well. One is unlikely to hit this
+-- on a 64-bit architecture, but on a 32-bit machine it's fairly possible to traverse
+-- @((0 :: 'Int') ...)@ far enough to encounter @0@ again.
 (...) :: Enum a => a -> Infinite a
 (...) = unsafeCycle . enumFrom
 {-# INLINE [0] (...) #-}
 
+infix 0 ...
+
 {-# RULES
 "ellipsis3Int" (...) = ellipsis3Int
 "ellipsis3Word" (...) = ellipsis3Word
@@ -248,7 +269,7 @@ ellipsis3Natural :: Natural -> Infinite Natural
 ellipsis3Natural = iterate' (+ 1)
 {-# INLINE ellipsis3Natural #-}
 
--- | Generate infinite sequences, starting from given elements,
+-- | Generate an infinite arithmetic progression, starting from given elements,
 -- similar to @[x,y..]@.
 -- For better user experience consider enabling @{\-# LANGUAGE PostfixOperators #-\}@:
 --
@@ -261,10 +282,16 @@ ellipsis3Natural = iterate' (+ 1)
 -- >>> :set -XPostfixOperators
 -- >>> Data.List.Infinite.take 10 ((EQ,GT)....)
 -- [EQ,GT,EQ,GT,EQ,GT,EQ,GT,EQ,GT]
+--
+-- Remember that 'Int' is a finite type as well: for a sufficiently large
+-- step of progression @y - x@ one may observe @((x :: Int, y)....)@ cycling back
+-- to emit @x@ fairly soon.
 (....) :: Enum a => (a, a) -> Infinite a
 (....) = unsafeCycle . uncurry enumFromThen
 {-# INLINE [0] (....) #-}
 
+infix 0 ....
+
 {-# RULES
 "ellipsis4Int" (....) = ellipsis4Int
 "ellipsis4Word" (....) = ellipsis4Word
@@ -322,15 +349,19 @@ instance Applicative Infinite where
 
 -- | 'Control.Applicative.ZipList' cannot be made a lawful 'Monad',
 -- but 'Infinite', being a
--- <https://hackage.haskell.org/package/adjunctions/docs/Data-Functor-Rep.html#t:Representable Representable>,
+-- [@Representable@](https://hackage.haskell.org/package/adjunctions/docs/Data-Functor-Rep.html#t:Representable),
 -- can. Namely, 'Control.Monad.join'
 -- picks up a diagonal of an infinite matrix of 'Infinite' ('Infinite' @a@).
--- This is mostly useful for parallel list comprehensions once
--- @{\-# LANGUAGE MonadComprehensions #-\}@ is enabled.
+-- Bear in mind that this instance gets slow
+-- very soon because of linear indexing, so it is not recommended to be used
+-- in practice.
 instance Monad Infinite where
   xs >>= f = go 0 xs
     where
-      go n (y :< ys) = f y !! n :< go (n + 1) ys
+      go !n (y :< ys) = (f y `index` n) :< go (n + 1) ys
+      index :: Infinite a -> Natural -> a
+      index ys n = head (genericDrop n ys)
+  {-# INLINE (>>=) #-}
   (>>) = (*>)
 
 -- | Get the first elements of an infinite list.
@@ -379,6 +410,10 @@ mapFB = (.)
   #-}
 
 -- | Flatten out an infinite list of non-empty lists.
+--
+-- The peculiar type with 'NonEmpty' is to guarantee that 'concat'
+-- is productive and results in an infinite list. Otherwise the
+-- concatenation of infinitely many @[a]@ could still be a finite list.
 concat :: Infinite (NonEmpty a) -> Infinite a
 concat = foldr (\(x :| xs) acc -> x :< (xs `prependList` acc))
 {-# NOINLINE [1] concat #-}
@@ -390,6 +425,10 @@ concat = foldr (\(x :| xs) acc -> x :< (xs `prependList` acc))
   #-}
 
 -- | First 'map' every element, then 'concat'.
+--
+-- The peculiar type with 'NonEmpty' is to guarantee that 'concatMap'
+-- is productive and results in an infinite list. Otherwise the
+-- concatenation of infinitely many @[b]@ could still be a finite list.
 concatMap :: (a -> NonEmpty b) -> Infinite a -> Infinite b
 concatMap f = foldr (\a acc -> let (x :| xs) = f a in x :< (xs `prependList` acc))
 {-# NOINLINE [1] concatMap #-}
@@ -417,6 +456,10 @@ intersperse a = foldr (\x -> (x :<) . (a :<))
 
 -- | Insert a non-empty list between adjacent elements of an infinite list,
 -- and subsequently flatten it out.
+--
+-- The peculiar type with 'NonEmpty' is to guarantee that 'intercalate'
+-- is productive and results in an infinite list. If separator is an empty list,
+-- concatenation of infinitely many @[a]@ could still be a finite list.
 intercalate :: NonEmpty a -> Infinite [a] -> Infinite a
 intercalate ~(a :| as) = foldr (\xs -> prependList xs . (a :<) . prependList as)
 {-# NOINLINE [1] intercalate #-}
@@ -430,7 +473,7 @@ intercalate ~(a :| as) = foldr (\xs -> prependList xs . (a :<) . prependList as)
 -- | Transpose rows and columns of an argument.
 --
 -- This is actually @distribute@ from
--- <https://hackage.haskell.org/package/distributive/docs/Data-Distributive.html#t:Distributive Distributive>
+-- [@Distributive@](https://hackage.haskell.org/package/distributive/docs/Data-Distributive.html#t:Distributive)
 -- type class in disguise.
 transpose :: Functor f => f (Infinite a) -> Infinite (f a)
 transpose xss = fmap head xss :< transpose (fmap tail xss)
@@ -441,9 +484,12 @@ subsequences = ([] :<) . map NE.toList . subsequences1
 
 -- | Generate an infinite list of all non-empty subsequences of the argument.
 subsequences1 :: Infinite a -> Infinite (NonEmpty a)
-subsequences1 (x :< xs) = (x :| []) :< foldr f (subsequences1 xs)
+subsequences1 = foldr go
   where
-    f ys r = ys :< (x `NE.cons` ys) :< r
+    go :: a -> Infinite (NonEmpty a) -> Infinite (NonEmpty a)
+    go x sxs = (x :| []) :< foldr f sxs
+      where
+        f ys r = ys :< (x `NE.cons` ys) :< r
 
 -- | Generate an infinite list of all permutations of the argument.
 permutations :: Infinite a -> Infinite (Infinite a)
@@ -461,12 +507,12 @@ permutations xs0 = xs0 :< perms xs0 []
           where
             (us, zs) = interleaveList' (f . (y :<)) ys r
 
--- |
+-- | Fold an infinite list from the left and return a list of successive reductions,
+-- starting from the initial accumulator:
+--
 -- > scanl f acc (x1 :< x2 :< ...) = acc :< f acc x1 :< f (f acc x1) x2 :< ...
 scanl :: (b -> a -> b) -> b -> Infinite a -> Infinite b
-scanl f = go
-  where
-    go z ~(x :< xs) = z :< go (f z x) xs
+scanl f z0 = (z0 :<) . flip (foldr (\x acc z -> let fzx = f z x in fzx :< acc fzx)) z0
 
 scanlFB :: (elt' -> elt -> elt') -> (elt' -> lst -> lst) -> elt -> (elt' -> lst) -> elt' -> lst
 scanlFB f cons = \elt g -> oneShot (\x -> let elt' = f x elt in elt' `cons` g elt')
@@ -486,9 +532,7 @@ scanlFB f cons = \elt g -> oneShot (\x -> let elt' = f x elt in elt' `cons` g el
 
 -- | Same as 'scanl', but strict in accumulator.
 scanl' :: (b -> a -> b) -> b -> Infinite a -> Infinite b
-scanl' f = go
-  where
-    go !z ~(x :< xs) = z :< go (f z x) xs
+scanl' f z0 = (z0 :<) . flip (foldr (\x acc z -> let !fzx = f z x in fzx :< acc fzx)) z0
 
 scanlFB' :: (elt' -> elt -> elt') -> (elt' -> lst -> lst) -> elt -> (elt' -> lst) -> elt' -> lst
 scanlFB' f cons = \elt g -> oneShot (\x -> let !elt' = f x elt in elt' `cons` g elt')
@@ -506,24 +550,25 @@ scanlFB' f cons = \elt g -> oneShot (\x -> let !elt' = f x elt in elt' `cons` g
     tail (scanl' f a bs)
   #-}
 
--- |
+-- | Fold an infinite list from the left and return a list of successive reductions,
+-- starting from the first element:
+--
 -- > scanl1 f (x0 :< x1 :< x2 :< ...) = x0 :< f x0 x1 :< f (f x0 x1) x2 :< ...
 scanl1 :: (a -> a -> a) -> Infinite a -> Infinite a
 scanl1 f (x :< xs) = scanl f x xs
 
--- | If you are looking how to traverse with a state, look no further:
+-- | Fold an infinite list from the left and return a list of successive reductions,
+-- keeping accumulator in a state:
 --
 -- > mapAccumL f acc0 (x1 :< x2 :< ...) =
 -- >   let (acc1, y1) = f acc0 x1 in
 -- >     let (acc2, y2) = f acc1 x2 in
 -- >       ...
 -- >         y1 :< y2 :< ...
+--
+-- If you are looking how to traverse with a state, look no further.
 mapAccumL :: (acc -> x -> (acc, y)) -> acc -> Infinite x -> Infinite y
-mapAccumL f = go
-  where
-    go s (x :< xs) = y :< go s' xs
-      where
-        (s', y) = f s x
+mapAccumL f = flip (foldr (\x acc s -> let (s', y) = f s x in y :< acc s'))
 
 mapAccumLFB :: (acc -> x -> (acc, y)) -> x -> (acc -> Infinite y) -> acc -> Infinite y
 mapAccumLFB f = \x r -> oneShot (\s -> let (s', y) = f s x in y :< r s')
@@ -604,6 +649,9 @@ repeatFB cons x = go
 -- | Repeat a non-empty list ad infinitum.
 -- If you were looking for something like @fromList :: [a] -> Infinite a@,
 -- look no further.
+--
+-- It would be less annoying to take @[a]@ instead of 'NonEmpty' @a@,
+-- but we strive to avoid partial functions.
 cycle :: NonEmpty a -> Infinite a
 cycle (x :| xs) = unsafeCycle (x : xs)
 {-# INLINE cycle #-}
@@ -628,6 +676,8 @@ unsafeCycleFB cons xs = go
   #-}
 
 -- | Build an infinite list from a seed value.
+--
+-- This is an anamorphism on infinite lists.
 unfoldr :: (b -> (a, b)) -> b -> Infinite a
 unfoldr f = go
   where
@@ -637,7 +687,7 @@ unfoldr f = go
 -- | Generate an infinite list of @f@ 0, @f@ 1, @f@ 2...
 --
 -- 'tabulate' and '(!!)' witness that 'Infinite' is
--- <https://hackage.haskell.org/package/adjunctions/docs/Data-Functor-Rep.html#t:Representable Representable>.
+-- [@Representable@](https://hackage.haskell.org/package/adjunctions/docs/Data-Functor-Rep.html#t:Representable).
 tabulate :: (Word -> a) -> Infinite a
 tabulate f = unfoldr (\n -> (f n, n + 1)) 0
 {-# INLINE tabulate #-}
@@ -645,70 +695,46 @@ tabulate f = unfoldr (\n -> (f n, n + 1)) 0
 -- | Take a prefix of given length.
 take :: Int -> Infinite a -> [a]
 take = GHC.Exts.inline genericTake
-
-takeFB :: (elt -> lst -> lst) -> lst -> elt -> (Int -> lst) -> Int -> lst
-takeFB cons nil x xs = \m -> if m <= 1 then x `cons` nil else x `cons` xs (m - 1)
-
 {-# INLINE [1] take #-}
 
-{-# INLINE [0] takeFB #-}
+{-# INLINE [1] genericTake #-}
+
+{-# INLINE [0] genericTakeFB #-}
 
 {-# RULES
-"take" [~1] forall n xs.
-  take n xs =
+"take"
+  take =
+    genericTake
+"genericTake" [~1] forall n xs.
+  genericTake n xs =
     GHC.Exts.build
       ( \cons nil ->
           if n >= 1
-            then foldr (takeFB cons nil) xs n
+            then foldr (genericTakeFB cons nil) xs n
             else nil
       )
-"takeList" [1] forall n xs.
-  foldr (takeFB (:) []) xs n =
-    take n xs
+"genericTakeList" [1] forall n xs.
+  foldr (genericTakeFB (:) []) xs n =
+    genericTake n xs
   #-}
 
 -- | Take a prefix of given length.
 genericTake :: Integral i => i -> Infinite a -> [a]
 genericTake n
   | n < 1 = const []
-  | otherwise = unsafeTake n
-  where
-    unsafeTake 1 (x :< _) = [x]
-    unsafeTake m (x :< xs) = x : unsafeTake (m - 1) xs
+  | otherwise = flip (foldr (\hd f m -> hd : (if m <= 1 then [] else f (m - 1)))) n
+
+genericTakeFB :: Integral i => (elt -> lst -> lst) -> lst -> elt -> (i -> lst) -> i -> lst
+genericTakeFB cons nil x xs = \m -> if m <= 1 then x `cons` nil else x `cons` xs (m - 1)
 
 -- | Drop a prefix of given length.
 drop :: Int -> Infinite a -> Infinite a
 drop = GHC.Exts.inline genericDrop
 
-dropFB :: (elt -> lst -> lst) -> elt -> (Int -> lst) -> Int -> lst
-dropFB cons x xs = \m -> if m < 1 then x `cons` xs m else xs (m - 1)
-
-{-# INLINE [1] drop #-}
-
-{-# INLINE [0] dropFB #-}
-
-{-# RULES
-"drop" [~1] forall n xs.
-  drop n xs =
-    build
-      ( \cons ->
-          if n >= 1
-            then foldr (dropFB cons) xs n
-            else foldr cons xs
-      )
-"dropList" [1] forall n xs.
-  foldr (dropFB (:<)) xs n =
-    drop n xs
-  #-}
-
 -- | Drop a prefix of given length.
 genericDrop :: Integral i => i -> Infinite a -> Infinite a
-genericDrop n
-  | n < 1 = id
-  | otherwise = unsafeDrop n
-  where
-    unsafeDrop 1 (_ :< xs) = xs
-    unsafeDrop m (_ :< xs) = unsafeDrop (m - 1) xs
+genericDrop = flip (para (\hd tl f m -> if m < 1 then hd :< tl else f (m - 1)))
+{-# INLINEABLE genericDrop #-}
 
 -- | Split an infinite list into a prefix of given length and the rest.
 splitAt :: Int -> Infinite a -> ([a], Infinite a)
@@ -718,18 +744,12 @@ splitAt = GHC.Exts.inline genericSplitAt
 genericSplitAt :: Integral i => i -> Infinite a -> ([a], Infinite a)
 genericSplitAt n
   | n < 1 = ([],)
-  | otherwise = unsafeSplitAt n
-  where
-    unsafeSplitAt 1 (x :< xs) = ([x], xs)
-    unsafeSplitAt m (x :< xs) = first (x :) (unsafeSplitAt (m - 1) xs)
+  | otherwise = flip (para (\hd tl f m -> if m <= 1 then ([hd], tl) else first (hd :) (f (m - 1)))) n
+{-# INLINEABLE genericSplitAt #-}
 
 -- | Take the longest prefix satisfying a predicate.
 takeWhile :: (a -> Bool) -> Infinite a -> [a]
-takeWhile p = go
-  where
-    go (x :< xs)
-      | p x = x : go xs
-      | otherwise = []
+takeWhile p = foldr (\x xs -> if p x then x : xs else [])
 
 takeWhileFB :: (elt -> Bool) -> (elt -> lst -> lst) -> lst -> elt -> lst -> lst
 takeWhileFB p cons nil = \x r -> if p x then x `cons` r else nil
@@ -752,27 +772,7 @@ takeWhileFB p cons nil = \x r -> if p x then x `cons` r else nil
 -- This function isn't productive (e. g., 'head' . 'dropWhile' @f@ won't terminate),
 -- if all elements of the input list satisfy the predicate.
 dropWhile :: (a -> Bool) -> Infinite a -> Infinite a
-dropWhile p = go
-  where
-    go xxs@(x :< xs)
-      | p x = go xs
-      | otherwise = xxs
-
-dropWhileFB :: (elt -> Bool) -> (elt -> lst -> lst) -> elt -> (Bool -> lst) -> (Bool -> lst)
-dropWhileFB p cons = \x r drp -> if drp && p x then r True else x `cons` r False
-
-{-# NOINLINE [1] dropWhile #-}
-
-{-# INLINE [0] dropWhileFB #-}
-
-{-# RULES
-"dropWhile" [~1] forall p xs.
-  dropWhile p xs =
-    build (\cons -> foldr (dropWhileFB p cons) xs True)
-"dropWhileList" [1] forall p xs.
-  foldr (dropWhileFB p (:<)) xs True =
-    dropWhile p xs
-  #-}
+dropWhile p = para (\x xs -> if p x then id else const (x :< xs))
 
 -- | Split an infinite list into the longest prefix satisfying a predicate and the rest.
 --
@@ -780,11 +780,7 @@ dropWhileFB p cons = \x r drp -> if drp && p x then r True else x `cons` r False
 -- (e. g., 'head' . 'snd' . 'span' @f@ won't terminate),
 -- if all elements of the input list satisfy the predicate.
 span :: (a -> Bool) -> Infinite a -> ([a], Infinite a)
-span p = go
-  where
-    go xxs@(x :< xs)
-      | p x = first (x :) (go xs)
-      | otherwise = ([], xxs)
+span p = para (\x xs -> if p x then first (x :) else const ([], x :< xs))
 
 -- | Split an infinite list into the longest prefix /not/ satisfying a predicate and the rest.
 --
@@ -797,10 +793,12 @@ break = span . (not .)
 -- | If a list is a prefix of an infinite list, strip it and return the rest.
 -- Otherwise return 'Nothing'.
 stripPrefix :: Eq a => [a] -> Infinite a -> Maybe (Infinite a)
-stripPrefix [] ys = Just ys
-stripPrefix (x : xs) (y :< ys)
-  | x == y = stripPrefix xs ys
-  | otherwise = Nothing
+stripPrefix [] = Just
+stripPrefix (p : ps) = flip (para alg) (p :| ps)
+  where
+    alg x xs acc (y :| ys)
+      | x == y = maybe (Just xs) acc (NE.nonEmpty ys)
+      | otherwise = Nothing
 
 -- | Group consecutive equal elements.
 group :: Eq a => Infinite a -> Infinite (NonEmpty a)
@@ -808,6 +806,9 @@ group = groupBy (==)
 
 -- | Overloaded version of 'group'.
 groupBy :: (a -> a -> Bool) -> Infinite a -> Infinite (NonEmpty a)
+-- Quite surprisingly, 'groupBy' is not a simple catamorphism.
+-- Since @f@ is not guaranteed to be transitive, it's a full-blown
+-- histomorphism, at which point a manual recursion becomes much more readable.
 groupBy f = go
   where
     go (x :< xs) = (x :| ys) :< go zs
@@ -846,10 +847,10 @@ tails = foldr (\x xss@(~(xs :< _)) -> (x :< xs) :< xss)
 
 -- | Check whether a list is a prefix of an infinite list.
 isPrefixOf :: Eq a => [a] -> Infinite a -> Bool
-isPrefixOf [] _ = True
-isPrefixOf (x : xs) (y :< ys)
-  | x == y = isPrefixOf xs ys
-  | otherwise = False
+isPrefixOf [] = const True
+isPrefixOf (p : ps) = flip (foldr alg) (p :| ps)
+  where
+    alg x acc (y :| ys) = x == y && maybe True acc (NE.nonEmpty ys)
 
 -- | Find the first pair, whose first component is equal to the first argument,
 -- and return the second component.
@@ -866,6 +867,13 @@ find f = foldr (\a a' -> if f a then a else a')
 --
 -- This function isn't productive (e. g., 'head' . 'filter' @f@ won't terminate),
 -- if no elements of the input list satisfy the predicate.
+--
+-- A common objection is that since it could happen that no elements of the input
+-- satisfy the predicate, the return type should be @[a]@ instead of 'Infinite' @a@.
+-- This would not however make 'filter' any more productive. Note that such
+-- hypothetical 'filter' could not ever generate @[]@ constructor, only @(:)@, so
+-- we would just have a more lax type gaining nothing instead. Same reasoning applies
+-- to other filtering \/ partitioning \/ searching functions.
 filter :: (a -> Bool) -> Infinite a -> Infinite a
 filter f = foldr (\a -> if f a then (a :<) else id)
 
@@ -906,13 +914,10 @@ partition f = foldr (\a -> if f a then first (a :<) else second (a :<))
 -- to avoid 'Prelude.error' on negative arguments.
 --
 -- This is actually @index@ from
--- <https://hackage.haskell.org/package/adjunctions/docs/Data-Functor-Rep.html#t:Representable Representable>
+-- [@Representable@](https://hackage.haskell.org/package/adjunctions/docs/Data-Functor-Rep.html#t:Representable)
 -- type class in disguise.
 (!!) :: Infinite a -> Word -> a
-(!!) = flip go
-  where
-    go 0 (x :< _) = x
-    go !m (_ :< ys) = go (m - 1) ys
+(!!) = foldr (\x acc m -> if m == 0 then x else acc (m - 1))
 
 infixl 9 !!
 
@@ -931,20 +936,14 @@ elemIndices = findIndices . (==)
 -- | Return an index of the first element, satisfying a predicate.
 -- If there is nothing to be found, this function will hang indefinitely.
 findIndex :: (a -> Bool) -> Infinite a -> Word
-findIndex f = go 0
-  where
-    go !n (x :< xs)
-      | f x = n
-      | otherwise = go (n + 1) xs
+findIndex f = flip (foldr (\x acc !m -> if f x then m else acc (m + 1))) 0
 
 -- | Return indices of all elements, satisfying a predicate.
 --
 -- This function isn't productive (e. g., 'head' . 'elemIndices' @f@ won't terminate),
 -- if no elements of the input list satisfy the predicate.
 findIndices :: (a -> Bool) -> Infinite a -> Infinite Word
-findIndices f = go 0
-  where
-    go !n (x :< xs) = (if f x then (n :<) else id) (go (n + 1) xs)
+findIndices f = flip (foldr (\x acc !m -> (if f x then (m :<) else id) (acc (m + 1)))) 0
 
 -- | Unzip an infinite list of tuples.
 unzip :: Infinite (a, b) -> (Infinite a, Infinite b)
@@ -976,30 +975,49 @@ unzip7 :: Infinite (a, b, c, d, e, f, g) -> (Infinite a, Infinite b, Infinite c,
 unzip7 = foldr (\(a, b, c, d, e, f, g) ~(as, bs, cs, ds, es, fs, gs) -> (a :< as, b :< bs, c :< cs, d :< ds, e :< es, f :< fs, g :< gs))
 {-# INLINE unzip7 #-}
 
--- | Split an infinite string into lines, by @\\n@.
+-- | Split an infinite string into lines, by @\\n@. Empty lines are preserved.
+--
+-- In contrast to their counterparts from "Data.List", it holds that
+-- 'unlines' @.@ 'lines' @=@ 'id'.
 lines :: Infinite Char -> Infinite [Char]
-lines xs = l :< lines xs'
+lines = foldr go
   where
-    (l, ~(_ :< xs')) = break (== '\n') xs
+    go '\n' xs = [] :< xs
+    go c ~(x :< xs) = (c : x) :< xs
 
 -- | Concatenate lines together with @\\n@.
+--
+-- In contrast to their counterparts from "Data.List", it holds that
+-- 'unlines' @.@ 'lines' @=@ 'id'.
 unlines :: Infinite [Char] -> Infinite Char
 unlines = foldr (\l xs -> l `prependList` ('\n' :< xs))
 
 -- | Split an infinite string into words, by any 'isSpace' symbol.
+-- Leading spaces are removed and, as underlined by the return type,
+-- repeated spaces are treated as a single delimiter.
 words :: Infinite Char -> Infinite (NonEmpty Char)
-words xs = (u :| us) :< words vs
+-- This is fundamentally a zygomorphism with 'isSpace' . 'head' as the small algebra.
+-- But manual implementation via catamorphism requires twice less calls of 'isSpace'.
+words = uncurry repack . foldr go
   where
-    u :< ys = dropWhile isSpace xs
-    (us, vs) = break isSpace ys
+    repack zs acc = maybe acc (:< acc) (NE.nonEmpty zs)
+
+    go x ~(zs, acc) = (zs', acc')
+      where
+        s = isSpace x
+        zs' = if s then [] else x : zs
+        acc' = if s then repack zs acc else acc
 
 wordsFB :: (NonEmpty Char -> lst -> lst) -> Infinite Char -> lst
-wordsFB cons = go
+wordsFB cons = uncurry repack . foldr go
   where
-    go xs = (u :| us) `cons` go vs
+    repack zs acc = maybe acc (`cons` acc) (NE.nonEmpty zs)
+
+    go x ~(zs, acc) = (zs', acc')
       where
-        u :< ys = dropWhile isSpace xs
-        (us, vs) = break isSpace ys
+        s = isSpace x
+        zs' = if s then [] else x : zs
+        acc' = if s then repack zs acc else acc
 
 {-# NOINLINE [1] words #-}
 
@@ -1011,6 +1029,10 @@ wordsFB cons = go
   #-}
 
 -- | Concatenate words together with a space.
+--
+-- The function is meant to be a counterpart of with 'words'.
+-- If you need to concatenate together 'Infinite' @[@'Char'@]@,
+-- use 'intercalate' @(@'pure' @' ')@.
 unwords :: Infinite (NonEmpty Char) -> Infinite Char
 unwords = foldr (\(l :| ls) acc -> l :< ls `prependList` (' ' :< acc))
 
@@ -1032,14 +1054,7 @@ nub = nubBy (==)
 
 -- | Overloaded version of 'nub'.
 nubBy :: (a -> a -> Bool) -> Infinite a -> Infinite a
-nubBy eq = go []
-  where
-    go seen (x :< xs)
-      | elemBy x seen = go seen xs
-      | otherwise = x :< go (x : seen) xs
-
-    elemBy _ [] = False
-    elemBy y (x : xs) = eq x y || elemBy y xs
+nubBy eq = flip (foldr (\x acc seen -> if List.any (`eq` x) seen then acc seen else x :< acc (x : seen))) []
 
 -- | Remove all occurrences of an element from an infinite list.
 delete :: Eq a => a -> Infinite a -> Infinite a
@@ -1047,11 +1062,7 @@ delete = deleteBy (==)
 
 -- | Overloaded version of 'delete'.
 deleteBy :: (a -> b -> Bool) -> a -> Infinite b -> Infinite b
-deleteBy eq x = go
-  where
-    go (y :< ys)
-      | eq x y = ys
-      | otherwise = y :< go ys
+deleteBy eq x = para (\y ys acc -> if eq x y then ys else y :< acc)
 
 -- | Take an infinite list and remove the first occurrence of every element
 -- of a finite list.
@@ -1079,11 +1090,7 @@ insert = insertBy compare
 
 -- | Overloaded version of 'insert'.
 insertBy :: (a -> a -> Ordering) -> a -> Infinite a -> Infinite a
-insertBy cmp x = go
-  where
-    go yys@(y :< ys) = case cmp x y of
-      GT -> y :< go ys
-      _ -> x :< yys
+insertBy cmp x = para (\y ys acc -> case cmp x y of GT -> y :< acc; _ -> x :< y :< ys)
 
 -- | Return all elements of an infinite list, which are simultaneously
 -- members of a finite list.
@@ -1097,3 +1104,42 @@ intersectBy eq xs ys = filter (\x -> List.any (eq x) ys) xs
 -- | Prepend a list to an infinite list.
 prependList :: [a] -> Infinite a -> Infinite a
 prependList = flip (F.foldr (:<))
+
+-- | Apply a function to every element of an infinite list and collect 'Just' results.
+--
+-- This function isn't productive (e. g., 'head' . 'mapMaybe' @f@ won't terminate),
+-- if no elements of the input list result in 'Just'.
+--
+-- @since 0.1.1
+mapMaybe :: (a -> Maybe b) -> Infinite a -> Infinite b
+mapMaybe = foldr . (maybe id (:<) .)
+
+-- | Keep only 'Just' elements.
+--
+-- This function isn't productive (e. g., 'head' . 'catMaybes' won't terminate),
+-- if no elements of the input list are 'Just'.
+--
+-- @since 0.1.1
+catMaybes :: Infinite (Maybe a) -> Infinite a
+catMaybes = foldr (maybe id (:<))
+
+-- | Apply a function to every element of an infinite list and
+-- separate 'Data.Either.Left' and 'Data.Either.Right' results.
+--
+-- This function isn't productive (e. g., 'head' . 'Data.Tuple.fst' .
+-- 'mapEither' @f@ won't terminate),
+-- if no elements of the input list result in 'Data.Either.Left' or 'Data.Either.Right'.
+--
+-- @since 0.1.1
+mapEither :: (a -> Either b c) -> Infinite a -> (Infinite b, Infinite c)
+mapEither = foldr . (either (first . (:<)) (second . (:<)) .)
+
+-- | Separate 'Data.Either.Left' and 'Data.Either.Right' elements.
+--
+-- This function isn't productive (e. g., 'head' . 'Data.Tuple.fst' . 'partitionEithers'
+-- won't terminate),
+-- if no elements of the input list are 'Data.Either.Left' or 'Data.Either.Right'.
+--
+-- @since 0.1.1
+partitionEithers :: Infinite (Either a b) -> (Infinite a, Infinite b)
+partitionEithers = foldr (either (first . (:<)) (second . (:<)))
